open Ident
open Location

exception NonExhaustive of Types.descr
exception Constraint of Types.descr * Types.descr
exception ShouldHave of Types.descr * string
exception WrongLabel of Types.descr * label
exception UnboundId of id * bool
exception UnboundExtId of Types.CompUnit.t * id
exception ShouldHave2 of Types.descr * string * Types.descr
exception Error of string

exception Warning of string * Types.t

val warning: loc -> string -> unit
val error: loc -> string -> 'a

include Custom.T

val from_comp_unit: (Types.CompUnit.t -> t) ref

val empty_env: t
val get_ns_table : t -> Ns.table

val register_types : Types.CompUnit.t -> t -> unit

val enter_ns : U.t -> Ns.t -> t -> t
val enter_schema: U.t -> string -> t -> t
val find_schema: U.t -> t -> string
val enter_cu : U.t -> Types.CompUnit.t -> t -> t 
val find_cu  : U.t -> t -> Types.CompUnit.t

val enter_value: id -> Types.t -> t -> t
val enter_values: (id * Types.t) list -> t -> t
val find_value: id -> t -> Types.t
val iter_values: t -> (id -> Types.t -> unit) -> unit

val enter_type: id -> Types.t -> t -> t
val enter_types: (id * Types.t) list -> t -> t
val find_type: id -> t -> Types.t

val type_defs: t -> (id * Ast.ppat) list -> (id * Types.t) list
val typ: t -> Ast.ppat -> Types.Node.t
val pat: t -> Ast.ppat -> Patterns.node

val dump_types: Format.formatter -> t -> unit
val dump_type: Format.formatter -> t -> U.t -> unit
val dump_schema_type:
  Format.formatter -> t -> (Schema_types.component_kind * U.t * U.t) -> unit

val dump_ns: Format.formatter -> t -> unit
val set_ns_table_for_printer: t -> unit

val type_expr: 
  t -> Ast.pexpr -> Typed.texpr * Types.descr

val type_let_decl: t -> Ast.ppat -> Ast.pexpr -> 
  t * Typed.let_decl * (id * Types.t) list

val type_let_funs: t -> Ast.pexpr list -> 
  t * Typed.texpr list * (id * Types.t) list
  (* Assume that all the expressions are Abstractions *)




val flatten: (Types.t -> bool -> Types.t) -> (Types.t -> bool -> Types.t)

(** {2 Schema stuff} *)

val get_schema: string -> Schema_types.schema  (** lookup schema by uri *)

val get_schema_names: t -> U.t list  (** registered schema names *)

(* Operators *)

type type_fun = Types.t -> bool -> Types.t

val register_op: string -> int -> (type_fun list -> type_fun) -> unit
